from django.utils.translation import gettext as _
from wbcore.menus import ItemPermission, MenuItem
from wbcore.permissions.shortcuts import is_internal_user
from wbmailing.models.mailing_lists import MailingListSubscriberChangeRequest

MAILINGLIST_MENUITEM = MenuItem(
    label=_("Mailing Lists"),
    endpoint="wbmailing:mailinglist-list",
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user), permissions=["wbmailing.view_mailinglist"]
    ),
    add=MenuItem(
        label=_("Create Mailing List"),
        endpoint="wbmailing:mailinglist-list",
        permission=ItemPermission(
            method=lambda request: is_internal_user(request.user), permissions=["wbmailing.add_mailinglist"]
        ),
    ),
)

MAILINGLISTSUBSCRIPTIONCHANGEREQUEST_MENUITEM = MenuItem(
    label=_("Subscription Requests"),
    endpoint="wbmailing:mailinglistsubscriberchangerequest-list",
    endpoint_get_parameters={"status": MailingListSubscriberChangeRequest.Status.PENDING.value},
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user),
        permissions=["wbmailing.view_mailinglistsubscriberchangerequest"],
    ),
    add=MenuItem(
        label=_("Create Subscription Request"),
        endpoint="wbmailing:mailinglistsubscriberchangerequest-list",
        permission=ItemPermission(
            method=lambda request: is_internal_user(request.user),
            permissions=["wbmailing.add_mailinglistsubscriberchangerequest"],
        ),
    ),
)



from .mailing_lists import (
    MailingListButtonConfig,
    MailingListEmailContactThroughModelButtonConfig,
    MailingListSubcriptionRequestButtonConfig,
)
from .mails import (
    MailButtonConfig,
    MailStatusMassMailButtonConfig,
    MassMailButtonConfig,
)



from typing import TYPE_CHECKING

import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from django.db.models import QuerySet
from wbcore import viewsets
from wbmailing.models import Mail, MailEvent

if TYPE_CHECKING:
    from plotly.graph_objs._figure import Figure


class MailStatusBarChartViewSet(viewsets.ChartViewSet):
    IDENTIFIER = "wbmailing:mailstatus-barchart"
    queryset = MailEvent.objects.all()

    def get_queryset(self) -> QuerySet[Mail]:
        mass_mail_id = self.kwargs["mass_mail_id"]

        return super().get_queryset().filter(mail__mass_mail_id=mass_mail_id)

    def get_plotly(self, queryset: QuerySet[Mail]) -> "Figure":
        df = pd.DataFrame(queryset.values("event_type"))
        df["count"] = 1

        df = df.groupby(["event_type"])["count"].count().reset_index().sort_values(by="count")
        fig = px.bar(df, x="count", y="event_type", orientation="h")
        fig.update_layout(
            yaxis={"showticklabels": True, "title": None},
            xaxis={"title": "Number"},
            title={"text": "Mail Status", "x": 0.5},
        )
        return fig


class MailClickBarChartViewSet(viewsets.ChartViewSet):
    IDENTIFIER = "wbmailing:mailclick-barchart"
    queryset = MailEvent.objects.all()

    def get_queryset(self) -> QuerySet[MailEvent]:
        mass_mail_id = self.kwargs["mass_mail_id"]

        return super().get_queryset().filter(mail__mass_mail_id=mass_mail_id, event_type=MailEvent.EventType.CLICKED)

    def get_plotly(self, queryset: QuerySet[MailEvent]) -> "Figure":
        df = pd.DataFrame(queryset.values("click_url"))
        df["count"] = 1
        df = df.groupby(["click_url"])["count"].count().reset_index().sort_values(by="count")
        fig = go.Figure(layout=dict(template="plotly"))  # noqa THis is necessary to prevent some racing condition
        fig = px.bar(df, x="count", y="click_url", text="click_url", orientation="h")
        fig.update_layout(
            yaxis={"visible": False, "showticklabels": False},
            xaxis={"title": "Number of clicks"},
            title={"text": "Clicked URLs", "x": 0.5},
        )
        return fig


class AbstractRawDataChartViewSet(viewsets.ChartViewSet):
    queryset = MailEvent.objects.all()
    ANALYTICS_PROPERTY: str

    def get_queryset(self) -> QuerySet[MailEvent]:
        mass_mail_id = self.kwargs["mass_mail_id"]

        return super().get_queryset().filter(mail__mass_mail_id=mass_mail_id, event_type=MailEvent.EventType.OPENED)

    def get_plotly(self, queryset: QuerySet[MailEvent]) -> "Figure":
        df = pd.DataFrame(queryset.values(self.ANALYTICS_PROPERTY))
        df["count"] = 1
        df = df.groupby(self.ANALYTICS_PROPERTY)["count"].count().reset_index().sort_values(by="count")
        fig = go.Figure()
        if not df.empty:
            fig = px.bar(df, x=self.ANALYTICS_PROPERTY, y="count")
            fig.update_layout(
                xaxis={"showticklabels": True, "title": None},
                yaxis={"showticklabels": True, "title": "Number"},
            )
        return fig


class ClientsBarChartViewSet(AbstractRawDataChartViewSet):
    IDENTIFIER = "wbmailing:client-barchart"
    ANALYTICS_PROPERTY = "raw_data__Client__Name"

    def get_plotly(self, queryset: QuerySet[MailEvent]) -> "Figure":
        fig = super().get_plotly(queryset)
        fig.update_layout(title={"text": "Platforms", "x": 0.5})
        return fig


class CountryBarChartViewSet(AbstractRawDataChartViewSet):
    IDENTIFIER = "wbmailing:country-barchart"
    ANALYTICS_PROPERTY = "raw_data__Geo__Country"

    def get_plotly(self, queryset: QuerySet[MailEvent]) -> "Figure":
        fig = super().get_plotly(queryset)
        fig.update_layout(title={"text": "Countries", "x": 0.5})
        return fig


class RegionBarChartViewSet(AbstractRawDataChartViewSet):
    IDENTIFIER = "wbmailing:region-barchart"
    ANALYTICS_PROPERTY = "raw_data__Geo__Region"

    def get_plotly(self, queryset: QuerySet[MailEvent]) -> "Figure":
        fig = super().get_plotly(queryset)
        fig.update_layout(title={"text": "Regions", "x": 0.5})
        return fig



from django.apps import apps
from django.utils.translation import gettext_lazy as _
from wbcore import filters as wb_filters
from wbmailing import models


class MassMailFilterSet(wb_filters.FilterSet):
    is_factsheet_massmail = wb_filters.BooleanFilter(
        label=_("Is Factsheet Mass Mail"), method="boolean_is_factsheet_massmail", default=False
    )

    def boolean_is_factsheet_massmail(self, queryset, name, value):
        if apps.is_installed("wbreport"):
            if value is True:
                return queryset.filter(mailing_lists__reports__isnull=False).distinct()
            elif value is False:
                return queryset.filter(mailing_lists__reports__isnull=True).distinct()
        return queryset

    class Meta:
        model = models.MassMail
        fields = {
            "subject": ["exact", "icontains"],
            "from_email": ["exact", "icontains"],
            "template": ["exact"],
            "mailing_lists": ["exact"],
            "created": ["gte", "exact", "lte"],
            "creator": ["exact"],
            "status": ["exact"],
        }


class MailFilter(wb_filters.FilterSet):
    never_open = wb_filters.BooleanFilter(label=_("Never Opened"), method="boolean_never_open")

    event_type = wb_filters.MultipleChoiceFilter(
        label=_("Events"), method="filter_events", choices=models.MailEvent.EventType.choices
    )
    rejected_reason = wb_filters.MultipleChoiceFilter(
        label=_("Rejection Reasons"), method="filter_rejected_reason", choices=models.MailEvent.RejectReason.choices
    )
    is_massmail_mail = wb_filters.BooleanFilter(label=_("Mass Mail"), method="boolean_is_massmail")
    status = wb_filters.ChoiceFilter(label=_("Status"), choices=models.MailEvent.EventType.choices)

    class Meta:
        model = models.Mail
        fields = {
            "mass_mail": ["exact"],
            "from_email": ["exact"],
            "to_email": ["exact"],
            "created": ["gte", "exact", "lte"],
        }

    def boolean_is_massmail(self, queryset, name, value):
        if value is True:
            return queryset.filter(mass_mail__isnull=False)
        elif value is False:
            return queryset.filter(mass_mail__isnull=True)
        return queryset

    def boolean_never_open(self, queryset, name, value):
        if value:
            return queryset.exclude(events__event_type=models.MailEvent.EventType.OPENED)
        return queryset

    def filter_events(self, queryset, name, value):
        if value:
            return queryset.filter(events__event_type=value)
        return queryset

    def filter_rejected_reason(self, queryset, name, value):
        if value:
            return queryset.filter(events__reject_reason=value)
        return queryset



from django.contrib.messages import warning
from django.contrib.postgres.aggregates import StringAgg
from django.db.models import Count, Exists, F, OuterRef, Q, Subquery
from django.db.models.functions import Coalesce
from django.shortcuts import get_object_or_404, redirect, render
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from django.views.generic import View
from rest_framework.decorators import action
from rest_framework.response import Response
from wbcore import serializers as wb_serializers
from wbcore import viewsets
from wbcore.contrib.directory.models import Company, EmailContact, Entry
from wbmailing import models, serializers
from wbmailing.filters import (
    EmailContactMailingListFilterSet,
    MailingListEmailContactThroughModelModelFilterSet,
    MailingListFilterSet,
)
from wbmailing.models import (
    MailingListEmailContactThroughModel,
    MailingListSubscriberChangeRequest,
)
from wbmailing.viewsets.buttons import (
    MailingListButtonConfig,
    MailingListEmailContactThroughModelButtonConfig,
    MailingListSubcriptionRequestButtonConfig,
)
from wbmailing.viewsets.display import (
    EmailContactMailingListDisplayConfig,
    MailingListDisplayConfig,
    MailingListEntryDisplayConfig,
    MailingListSubscriberChangeRequestDisplayConfig,
    MailingListSubscriberRequestEntryDisplayConfig,
    MailingListSubscriberRequestMailingListDisplayConfig,
)
from wbmailing.viewsets.endpoints import (
    EmailContactMailingListEndpointConfig,
    MailingListEntryEndpointConfig,
    MailingListSubscriberRequestEntryEndpointConfig,
    MailingListSubscriberRequestMailingListEndpointConfig,
)
from wbmailing.viewsets.titles import (
    EmailContactMailingListTitleConfig,
    MailingListEntryTitleConfig,
    MailingListSubscriberChangeRequestTitleConfig,
    MailingListSubscriberRequestEntryTitleConfig,
    MailingListTitleConfig,
)


class MailingListRepresentationViewSet(viewsets.RepresentationViewSet):
    IDENTIFIER = "wbmailing:mailinglist"

    ordering_fields = ("title",)
    search_fields = ("title",)
    queryset = models.MailingList.objects.all()
    serializer_class = serializers.MailingListRepresentationSerializer
    filterset_class = MailingListFilterSet


class MailingListModelViewSet(viewsets.ModelViewSet):
    IDENTIFIER = "wbmailing:mailinglist"

    queryset = models.MailingList.objects.all()

    def get_serializer_class(self):
        if getattr(self, "action", None) == "list":
            return serializers.MailingListListSerializer
        return serializers.MailingListModelSerializer

    filterset_class = MailingListFilterSet

    ordering_fields = ordering = ("title",)
    search_fields = ("title",)

    display_config_class = MailingListDisplayConfig
    title_config_class = MailingListTitleConfig
    button_config_class = MailingListButtonConfig

    def get_queryset(self):
        qs = self.filter_queryset(
            models.MailingList.objects.annotate(
                nb_subscribers=Coalesce(
                    Subquery(
                        MailingListEmailContactThroughModel.objects.filter(
                            mailing_list=OuterRef("pk"), status=MailingListEmailContactThroughModel.Status.SUBSCRIBED
                        )
                        .values("mailing_list")
                        .annotate(c=Count("mailing_list"))
                        .values("c")[:1]
                    ),
                    0,
                )
            )
        )
        qs = qs.prefetch_related("email_contacts")
        return qs


class MailingListSubscriberChangeRequestModelViewSet(viewsets.ModelViewSet):
    queryset = (
        models.MailingListSubscriberChangeRequest.objects.select_related("email_contact")
        .select_related("mailing_list")
        .select_related("requester")
        .select_related("approver")
    ).annotate(entry_repr=F("email_contact__entry__computed_str"))

    serializer_class = serializers.MailingListSubscriberChangeRequestModelSerializer

    search_fields = ("email_contact__address", "mailing_list__title")
    ordering_fields = ("email_contact__address", "mailing_list__title", "created")
    ordering = ("-created",)
    filterset_fields = {
        "email_contact": ["exact"],
        "mailing_list": ["exact"],
        "requester": ["exact"],
        "created": ["gte", "exact", "lte"],
        "status": ["exact"],
    }

    display_config_class = MailingListSubscriberChangeRequestDisplayConfig
    title_config_class = MailingListSubscriberChangeRequestTitleConfig
    button_config_class = MailingListSubcriptionRequestButtonConfig

    def add_messages(self, request, instance=None, **kwargs):
        if instance:
            _type = (
                "subscribed"
                if instance.type == MailingListSubscriberChangeRequest.Type.SUBSCRIBING
                else "unsubscribed"
            )
            warning(
                request,
                f"Upon approval, This change request will {_type} {instance.email_contact.address} to {instance.mailing_list.title}",
            )

    @action(detail=False, methods=["GET"])
    def approveall(self, request, pk=None):
        for request in models.MailingListSubscriberChangeRequest.objects.filter(
            status=models.MailingListSubscriberChangeRequest.Status.PENDING
        ).all():
            request.approve()
            request.save()
        return Response({"send": True})


class MailingListSubscriberRequestMailingListModelViewSet(MailingListSubscriberChangeRequestModelViewSet):
    IDENTIFIER = "wbmailing:mailing_list-mailinglistsubscriberchangerequest"

    display_config_class = MailingListSubscriberRequestMailingListDisplayConfig
    endpoint_config_class = MailingListSubscriberRequestMailingListEndpointConfig

    def get_queryset(self):
        return super().get_queryset().filter(mailing_list__id=self.kwargs["mailing_list_id"])


class MailingListEmailContactThroughModelModelViewSet(viewsets.ModelViewSet):
    queryset = MailingListEmailContactThroughModel.objects.select_related(
        "email_contact",
        "mailing_list",
    )
    search_fields = ["email_contact__address", "email_contact__entry__computed_str"]
    button_config_class = MailingListEmailContactThroughModelButtonConfig
    serializer_class = serializers.MailingListEmailContactThroughModelModelSerializer

    ordering = ["email_contact__address", "mailing_list__title"]
    ordering_fields = ["email_contact__address", "status", "mailing_list__title"]

    filterset_class = MailingListEmailContactThroughModelModelFilterSet

    @action(detail=True, methods=["GET", "PATCH"])
    def removeexpirationdate(self, request, mailing_list_id=None, pk=None, **kwargs):
        through = get_object_or_404(MailingListEmailContactThroughModel, id=pk)
        if (qs := through.requests.filter(expiration_date__isnull=False)).exists():
            last_request = qs.latest("created")
            last_request.expiration_date = None
            last_request.save()
        return Response({"send": True})

    @action(detail=True, methods=["GET", "PATCH"])
    def delete(self, request, mailing_list_id=None, pk=None, **kwargs):
        through = get_object_or_404(MailingListEmailContactThroughModel, id=pk)
        through.delete()
        return Response({"send": True})

    @action(detail=True, methods=["GET", "PATCH"])
    def unsubscribe(self, request, pk=None, **kwargs):
        through = get_object_or_404(MailingListEmailContactThroughModel, id=pk)
        if through.status == MailingListEmailContactThroughModel.Status.SUBSCRIBED:
            through.change_state(
                reason=_("Unsubscribed by {}").format(str(request.user)),
                requester=request.user.profile,
                approver=request.user.profile,
                automatically_approve=True,
            )
        return Response({"send": True})

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .annotate(
                expiration_date=MailingListEmailContactThroughModel.get_expired_date_subquery(),
                in_charge=StringAgg(F("email_contact__entry__relationship_managers__computed_str"), delimiter=", "),
                is_pending_request_change=Exists(
                    MailingListSubscriberChangeRequest.objects.filter(
                        relationship=OuterRef("pk"), status=MailingListSubscriberChangeRequest.Status.PENDING
                    )
                ),
                is_public=F("mailing_list__is_public"),
            )
        )


class EmailContactMailingListModelViewSet(MailingListEmailContactThroughModelModelViewSet):
    filterset_class = EmailContactMailingListFilterSet

    display_config_class = EmailContactMailingListDisplayConfig
    title_config_class = EmailContactMailingListTitleConfig
    endpoint_config_class = EmailContactMailingListEndpointConfig

    def get_queryset(self):
        return super().get_queryset().filter(mailing_list=self.kwargs["mailing_list_id"])


class MailingListEntryModelViewSet(MailingListEmailContactThroughModelModelViewSet):
    filterset_fields = {"mailing_list": ["exact"], "status": ["exact"]}

    display_config_class = MailingListEntryDisplayConfig
    title_config_class = MailingListEntryTitleConfig
    endpoint_config_class = MailingListEntryEndpointConfig

    @cached_property
    def casted_entry(self):
        return get_object_or_404(Entry, id=self.kwargs["entry_id"]).get_casted_entry()

    @cached_property
    def primary_email(self):
        return EmailContact.objects.filter(entry=self.kwargs["entry_id"], primary=True).first()

    def add_messages(self, request, instance=None, **kwargs):
        if not self.primary_email:
            warning(
                request,
                "This person does not have a primary email. Adds one first before adding them to a mailing list.",
            )

    def get_queryset(self):
        qs = super().get_queryset()
        entry = self.casted_entry
        if isinstance(entry, Company):
            email_contacts = EmailContact.objects.filter(
                Q(entry=entry) | Q(entry__in=entry.employees.all())
            ).distinct()
            qs = qs.filter(email_contact__in=email_contacts)
        else:
            qs = qs.filter(email_contact__entry=entry)
        return qs


class MailingListSubscriberRequestEntryModelViewSet(MailingListSubscriberChangeRequestModelViewSet):
    IDENTIFIER = "wbmailing:entry-mailinglistsubscriberchangerequest"
    display_config_class = MailingListSubscriberRequestEntryDisplayConfig
    title_config_class = MailingListSubscriberRequestEntryTitleConfig
    endpoint_config_class = MailingListSubscriberRequestEntryEndpointConfig

    @cached_property
    def primary_email(self):
        return EmailContact.objects.filter(entry=self.kwargs["entry_id"], primary=True).first()

    def get_serializer_class(self):
        if self.primary_email:

            class Serializer(serializers.MailingListSubscriberChangeRequestModelSerializer):
                email_contact = wb_serializers.PrimaryKeyRelatedField(
                    queryset=EmailContact.objects.filter(entry=self.kwargs["entry_id"]),
                    label=_("Email"),
                    many=False,
                    default=self.primary_email,
                )

        else:

            class Serializer(serializers.MailingListSubscriberChangeRequestModelSerializer):
                email_contact = wb_serializers.PrimaryKeyRelatedField(label=_("No Emails"), read_only=True)

        return Serializer

    def get_queryset(self):
        return super().get_queryset().filter(email_contact__entry__id=self.kwargs["entry_id"])


#################
# TODO Old system
#################


class ManageMailingListSubscriptions(View):
    def get(self, request, email_contact_id, *args, **kwargs):
        email_contact = get_object_or_404(EmailContact, id=email_contact_id)
        context = {
            "title": _("Manage Mailing List Subscriptions"),
            "email_contact": email_contact,
            "mailing_lists": models.MailingList.get_subscribed_mailing_lists(email_contact),
        }
        return render(request, "mailing/manage_mailing_list_subscriptions.html", context=context)


class UnsubscribeView(View):
    def get(self, request, email_contact_id, mailing_list_id, *args, **kwargs):
        email_contact = get_object_or_404(EmailContact, id=email_contact_id)
        mailing_list = get_object_or_404(models.MailingList, id=mailing_list_id)
        mailing_list.unsubscribe(
            email_contact, reason=_("The user requested to be unsubscribed."), automatically_approve=True
        )
        return redirect(
            "wbmailing:manage_mailing_list_subscriptions",
            email_contact_id=email_contact_id,
        )



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig
from wbmailing.models import Mail


class MailStatusMassMailEndpointConfig(EndpointViewConfig):
    PK_FIELD = "mail_id"

    def get_endpoint(self, **kwargs):
        return None

    def get_list_endpoint(self, **kwargs):
        return reverse(
            "wbmailing:massmail-mailstatus-list",
            args=[self.view.kwargs["mass_mail_id"]],
            request=self.request,
        )

    def get_instance_endpoint(self, **kwargs):
        endpoint = reverse(
            "wbmailing:mail-list",
        )
        if self.instance and (mass_mail_id := self.view.kwargs.get("mass_mail_id", None)):
            obj = self.view.get_object()
            mail = Mail.objects.filter(mass_mail=mass_mail_id, to_email=obj).latest("created")
            return f"{endpoint}{mail.id}"
        return endpoint


class MailEventMailEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return None

    def get_list_endpoint(self, **kwargs):
        return reverse(
            "wbmailing:mail-mailevent-list",
            args=[self.view.kwargs["mail_id"]],
            request=self.request,
        )


class MailEventMassMailEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return None

    def get_list_endpoint(self, **kwargs):
        return reverse(
            "wbmailing:massmail-mailevent-list",
            args=[self.view.kwargs["mass_mail_id"]],
            request=self.request,
        )



from unittest.mock import patch

import pytest
from anymail.signals import AnymailTrackingEvent, tracking
from django.core import mail
from django.test import override_settings
from rest_framework.test import APIRequestFactory
from wbcore.contrib.directory.factories import EmailContactFactory
from wbcore.contrib.documents.factories import DocumentFactory
from wbcore.test.utils import get_or_create_superuser
from wbmailing.backend import SendgridEmailBackend
from wbmailing.factories import ToEmailMailFactory
from wbmailing.models import MailEvent, MassMail
from wbmailing.models.mails import (
    can_administrate_mail,
    send_mail_task,
    send_mass_mail_as_task,
)


@pytest.mark.django_db
class TestSpecificModels:
    @patch("wbmailing.models.mails.send_mass_mail_as_task.delay")
    def test_send_mail_task(self, send_mass_mail_as_task, mass_mail_factory, mailing_list_factory):
        ec = EmailContactFactory()
        ml = mailing_list_factory(email_contacts=(ec,))
        mm = mass_mail_factory(mailing_lists=(ml,))

        send_mail_task(mm.id)
        send_mass_mail_as_task.assert_called()
        assert send_mass_mail_as_task.call_count == 1

    def test_send_mass_mail_as_task(self, mass_mail_factory, mailing_list_factory):
        ec = EmailContactFactory()
        ml = mailing_list_factory(email_contacts=(ec,))
        mass_mail = mass_mail_factory(mailing_lists=(ml,))
        assert len(mail.outbox) == 0
        for subscriber in mass_mail.get_mail_addresses():
            send_mass_mail_as_task(mass_mail.id, subscriber["address"])
        assert len(mail.outbox) == 1

    @pytest.mark.parametrize("status, expected", [("PENDING", "DENIED")])
    def test_deny(self, mailing_list_subscriber_change_request_factory, status, expected):
        mlscr = mailing_list_subscriber_change_request_factory()
        mlscr.deny(description="SPAM")
        assert mlscr.status == expected

    @pytest.mark.parametrize("status, expected", [("DRAFT", "PENDING")])
    def test_submit(self, mass_mail_factory, status, expected):
        ml = mass_mail_factory()
        ml.submit()
        assert ml.status == expected

    @pytest.mark.parametrize("status, expected", [("PENDING", "DENIED")])
    def test_deny2(self, mass_mail_factory, status, expected):
        ml = mass_mail_factory(status=MassMail.Status.PENDING)
        ml.deny()
        assert ml.status == expected

    @pytest.mark.parametrize("status, expected", [("PENDING", "DRAFT")])
    def test_revise(self, mass_mail_factory, status, expected):
        ml = mass_mail_factory(status=MassMail.Status.PENDING)
        ml.revise()
        assert ml.status == expected

    @patch("wbmailing.models.mails.send_mail_task.delay")
    def test_send(self, send_mail_task, mass_mail_factory):
        mm = mass_mail_factory(status=MassMail.Status.PENDING)
        mm.send()
        send_mail_task.assert_called()
        assert send_mail_task.call_count == 1

    def test_create_email(self, mass_mail_factory):
        request = APIRequestFactory().get("")
        request.user = get_or_create_superuser()
        mm = mass_mail_factory(attachments=(DocumentFactory(),))
        EmailContactFactory(address=request.user.email)
        mm.template = None
        mm.save()
        msg = mm.create_email(request.user.email)
        assert msg

    @patch("wbmailing.models.mails.send_mail_as_task.delay")
    def test_resend_no_massmail(self, send_mail_as_task):
        request = APIRequestFactory().get("")
        request.user = get_or_create_superuser()
        obj = ToEmailMailFactory(attachments=(DocumentFactory(),))
        obj.mass_mail = None
        obj.save()
        obj.resend()
        send_mail_as_task.assert_called()
        assert send_mail_as_task.call_count == 1

    @patch("wbmailing.models.mails.send_mail_as_task.delay")
    def test_resend_no_template(self, send_mail_as_task):
        request = APIRequestFactory().get("")
        request.user = get_or_create_superuser()
        obj = ToEmailMailFactory()
        obj.mass_mail.template = None
        obj.mass_mail.save()
        obj.resend()
        send_mail_as_task.assert_called()
        assert send_mail_as_task.call_count == 1

    @override_settings(EMAIL_BACKEND="anymail.backends.test.EmailBackend")
    @pytest.mark.parametrize(
        "exits_message_id, event_type, reject_reason",
        [
            ("YES", MailEvent.EventType.SENT, None),
            ("YES", MailEvent.EventType.SENT, MailEvent.RejectReason.SPAM),
            ("YES", "Other", "Unknow"),
            ("NO", "Other", None),
        ],
    )
    def test_handle_mail_tracking(self, exits_message_id, event_type, reject_reason, mail_factory, mail_event_factory):
        request = APIRequestFactory().get("")
        request.user = get_or_create_superuser()
        ml = ToEmailMailFactory()
        nb_mail_send = len(mail.outbox)
        esp_event = {}
        if reject_reason:
            mailevent = mail_event_factory(mail=ml, description="", reject_reason=reject_reason)
            if reject_reason == MailEvent.RejectReason.SPAM:
                esp_event["reason"] = "SPAM"
            else:
                esp_event["response"] = "Unknow"
        else:
            mailevent = mail_event_factory(mail=ml)

        to = list(ml.to_email.values_list("address", flat=True))
        msg = ml.get_mailmessage(ml.subject, ml.body, request.user.email, to, attachments=ml.documents.all())
        msg.send()
        assert len(mail.outbox) == nb_mail_send + 1
        assert msg.anymail_status.status == {"sent"}
        if exits_message_id == "YES":
            message_id = mailevent.mail.message_ids[0]
        else:
            message_id = msg.anymail_status.message_id

        esp_event["ip"] = mailevent.ip
        esp_event["useragent"] = mailevent.user_agent
        mevent = AnymailTrackingEvent(
            event_type=event_type,
            message_id=message_id,
            timestamp=mailevent.timestamp,
            event_id=mailevent.id,
            recipient=mailevent.recipient,
            reject_reason=mailevent.reject_reason,
            description=mailevent.description,
            user_agent=mailevent.user_agent,
            click_url=mailevent.click_url,
            esp_event=esp_event,
        )
        # sender(class) – The source of the event.  (One of theanymail.webhook.*Viewclasses, but you generally won’t examine this parameter; it’s required by Django’s signalmechanism.
        tracking.send(sender=msg, event=mevent, esp_name="SendGrid")

        if exits_message_id == "YES":
            assert MailEvent.objects.filter(mail=ml, recipient=mailevent.recipient).count() == 2
        else:
            assert MailEvent.objects.filter(mail=ml, recipient=mailevent.recipient).count() == 1

    @override_settings(EMAIL_BACKEND="anymail.backends.test.EmailBackend", ANYMAIL_SENDGRID_API_KEY="TEST")
    @patch("wbmailing.backend.SendgridEmailBackend._send")
    @pytest.mark.parametrize("resend, exit_document", [(False, False), (False, True), (True, False)])
    def test_SendgridEmailBackend_send_messages(self, mock_send, resend, exit_document):
        request = APIRequestFactory().get("")
        request.user = get_or_create_superuser()
        num_sent = SendgridEmailBackend().send_messages(None)
        assert num_sent == 0
        doc1 = DocumentFactory()
        ml = ToEmailMailFactory(cc_email=(EmailContactFactory(),), attachments=(doc1,))
        to = list(ml.to_email.values_list("address", flat=True))
        cc = list(ml.cc_email.values_list("address", flat=True))
        bcc = ["lemissan@atonra.ch"]
        if resend:
            msg = ml.get_mailmessage(
                ml.subject, ml.body, request.user.email, to, attachments=ml.documents.all(), mail=ml
            )
        else:
            msg = ml.get_mailmessage(
                ml.subject, ml.body, request.user.email, to, bcc=bcc, cc=cc, attachments=ml.documents.all()
            )
        msg.send()

        if exit_document:
            for name, payload, mimetype in msg.attachments:
                doc1.name = name
                doc1.save()

        email_messages = [msg]

        mock_send.return_value.status_code = 200
        mock_send.return_value.json.return_value = msg

        num_sent = SendgridEmailBackend().send_messages(email_messages)
        mock_send.assert_called()
        assert mock_send.call_count == 1
        assert num_sent == 1

    def test_can_administrate_mail(self, mass_mail_factory):
        request = APIRequestFactory().get("")
        request.user = get_or_create_superuser()
        obj = mass_mail_factory()
        result = can_administrate_mail(obj, request.user)
        assert result



from datetime import datetime, timedelta
from unittest.mock import patch

import pytest
from django.db.models import Q
from django.utils import timezone
from faker import Faker
from wbmailing.factories import MailingListSubscriberChangeRequestFactory
from wbmailing.models import MassMail
from wbmailing.models.mailing_lists import MailingListEmailContactThroughModel
from wbmailing.tasks import (
    check_and_remove_expired_mailinglist_subscription,
    periodic_send_mass_mail_as_tasks,
)

fake = Faker()


@pytest.mark.django_db
class TestSpecificTasks:
    @pytest.mark.parametrize("expiration_date", [fake.past_date()])
    def test_check_and_remove_expired_mailinglist_subscription(self, expiration_date, mailing_list, email_contact):
        request = MailingListSubscriberChangeRequestFactory(
            expiration_date=expiration_date, email_contact=email_contact, mailing_list=mailing_list
        )
        request.approve()
        request.save()
        assert request.relationship.status == MailingListEmailContactThroughModel.Status.SUBSCRIBED

        check_and_remove_expired_mailinglist_subscription()
        request.refresh_from_db()
        assert request.relationship.status == MailingListEmailContactThroughModel.Status.UNSUBSCRIBED
        request.refresh_from_db()
        assert request.expiration_date is None

    @patch("wbmailing.models.mails.send_mail_task.delay")
    def test_periodic_tasks(self, send_mail_task, mass_mail_factory):
        mass_mail_factory(status=MassMail.Status.PENDING, send_at=None)
        send_date = timezone.now() - timedelta(minutes=15)
        mass_mail_factory(status=MassMail.Status.SEND_LATER, send_at=send_date)
        assert (
            MassMail.objects.filter(Q(status=MassMail.Status.SEND_LATER) & Q(send_at__lte=datetime.now())).count() == 1
        )
        periodic_send_mass_mail_as_tasks()
        send_mail_task.assert_called()
        assert send_mail_task.call_count == 1
        assert (
            MassMail.objects.filter(Q(status=MassMail.Status.SEND_LATER) & Q(send_at__lte=datetime.now())).count() == 0
        )



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig


class MailingListSubscriberRequestMailingListEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbmailing:mailing_list-mailinglistsubscriberchangerequest-list",
            args=[self.view.kwargs["mailing_list_id"]],
            request=self.request,
        )


class EmailContactMailingListEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return None

    def get_list_endpoint(self, **kwargs):
        return reverse(
            "wbmailing:mailing_list-email_contacts-list",
            args=[self.view.kwargs["mailing_list_id"]],
            request=self.request,
        )

    def get_create_endpoint(self, **kwargs):
        return reverse(
            "wbmailing:mailing_list-mailinglistsubscriberchangerequest-list",
            args=[self.view.kwargs["mailing_list_id"]],
            request=self.request,
        )


class MailingListEntryEndpointConfig(EndpointViewConfig):
    PK_FIELD = "mailing_list"

    def get_list_endpoint(self, **kwargs):
        return reverse(
            "wbmailing:entry-mailinglist-list",
            args=[self.view.kwargs["entry_id"]],
            request=self.request,
        )

    def get_instance_endpoint(self, **kwargs):
        return reverse(
            "wbmailing:mailinglist-list",
            args=[],
            request=self.request,
        )

    def get_delete_endpoint(self, **kwargs):
        return None

    def get_create_endpoint(self, **kwargs):
        if getattr(self.view, "primary_email", None):
            return reverse(
                "wbmailing:entry-mailinglistsubscriberchangerequest-list",
                args=[self.view.kwargs["entry_id"]],
                request=self.request,
            )
        return None


class MailMailingListChartEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbmailing:mailing_list-maileventchart-list",
            args=[self.view.kwargs["mailing_list_id"]],
            request=self.request,
        )


class MailingListSubscriberRequestEntryEndpointConfig(EndpointViewConfig):
    def get_instance_endpoint(self, **kwargs):
        if getattr(self.view, "primary_email", None):
            return super().get_instance_endpoint(**kwargs)
        return None

    def get_create_endpoint(self, **kwargs):
        return self.get_instance_endpoint(**kwargs)

    def get_list_endpoint(self, **kwargs):
        return reverse(
            "wbmailing:entry-mailinglistsubscriberchangerequest-list",
            args=[self.view.kwargs["entry_id"]],
            request=self.request,
        )



import pytest
from django.core import mail
from rest_framework import status
from rest_framework.test import APIRequestFactory
from wbcore.contrib.directory.factories import EmailContactFactory
from wbcore.test.utils import get_kwargs
from wbmailing.factories import CustomMassMailFactory
from wbmailing.models import MassMail
from wbmailing.viewsets.mails import MassMailModelViewSet


@pytest.mark.django_db
class TestSpecificSerializers:
    @pytest.mark.parametrize("mvs, factory", [(MassMailModelViewSet, CustomMassMailFactory)])
    def test_serializers_mail(self, mvs, factory, user):
        request = APIRequestFactory().get("")
        request.user = user
        obj = factory(status=MassMail.Status.SENT)
        EmailContactFactory(address=request.user.email)
        nb_mail_send = len(mail.outbox)
        request.query_params = {}
        mvs.request = request
        mvs.kwargs = get_kwargs(obj, mvs, request)
        mvs.kwargs["pk"] = obj.pk
        response = mvs().sendtestmail(request)
        assert response.status_code == status.HTTP_200_OK
        assert response.data
        assert len(mail.outbox) == nb_mail_send + 1



from datetime import timedelta

from anymail.backends.mailgun import EmailBackend as AnymailMailgunBackend
from anymail.backends.mailjet import EmailBackend as AnymailMailjetBackend
from anymail.backends.mandrill import EmailBackend as AnymailMandrillBackend
from anymail.backends.postmark import EmailBackend as AnymailPostmarkBackend
from anymail.backends.sendgrid import EmailBackend as AnymailSendgridBackend
from anymail.backends.sendinblue import EmailBackend as AnymailSendinblueBackend
from anymail.exceptions import AnymailError
from django.conf import settings
from django.core.mail.backends.console import EmailBackend as ConsoleBackend
from django.utils import timezone
from sentry_sdk import capture_message
from wbcore.utils.html import convert_html2text
from wbmailing.models import Mail, MailEvent


class SendMessagesMixin:
    def _process_msg(self, message):
        mail = message.mail if hasattr(message, "mail") else None
        mass_mail = message.mass_mail if hasattr(message, "mass_mail") else None
        if mail:
            event_type = MailEvent.EventType.RESENT
        else:
            if not hasattr(message, "silent_mail") or (hasattr(message, "silent_mail") and not message.silent_mail):
                mail = Mail.create_mail_from_mailmessage(message, user=getattr(message, "user", None))
            event_type = MailEvent.EventType.CREATED
        if mass_mail:
            message.tags = [f"massmail-{mass_mail.id}"]
        else:
            message.tags = [f"mail-{mail.id}"]
        if mail and mail.body:
            # We reset the body text and html field with what might have been computed in create_mail_from_mailmessage
            message.body = convert_html2text(mail.body)
            message.alternatives = []
            message.attach_alternative(mail.body, "text/html")
        return event_type, mail

    def send_messages(self, email_messages):
        """
        Sends one or more EmailMessage objects and returns the number of email
        messages sent.
        """
        # This API is specified by Django's core BaseEmailBackend
        # (so you can't change it to, e.g., return detailed status).
        # Subclasses shouldn't need to override.
        from wbmailing.models import MailEvent

        num_sent = 0
        if not email_messages:
            return num_sent

        created_session = self.open()

        try:
            for message in email_messages:
                try:
                    event_type, mail = self._process_msg(message)
                    sent = self._send(message)
                    if mail:
                        MailEvent.objects.create(
                            mail=mail, event_type=event_type, timestamp=timezone.now() - timedelta(seconds=1)
                        )
                        mail.message_ids.append(message.anymail_status.message_id)
                        mail.last_send = timezone.now()
                        mail.save()
                except AnymailError as e:
                    capture_message(e)
                    if self.fail_silently:
                        sent = False
                    else:
                        raise
                if sent:
                    num_sent += 1
        finally:
            if created_session:
                self.close()

        return num_sent


class PostmarkEmailBackend(SendMessagesMixin, AnymailPostmarkBackend):
    def send_messages(self, email_messages):
        if settings.WBMAILING_POSTMARK_BROADCAST_STREAM_ID:
            for msg in email_messages:
                if hasattr(msg, "mass_mail") and msg.mass_mail:
                    msg.esp_extra = {"MessageStream": settings.WBMAILING_POSTMARK_BROADCAST_STREAM_ID}

        return super().send_messages(email_messages)


class SendgridEmailBackend(SendMessagesMixin, AnymailSendgridBackend):
    pass


class MailgunEmailBackend(SendMessagesMixin, AnymailMailgunBackend):
    pass


class MailjetEmailBackend(SendMessagesMixin, AnymailMailjetBackend):
    pass


class MandrillEmailBackend(SendMessagesMixin, AnymailMandrillBackend):
    pass


class SendinblueEmailBackend(SendMessagesMixin, AnymailSendinblueBackend):
    pass


class ConsoleEmailBackend(SendMessagesMixin, ConsoleBackend):
    def send_messages(self, email_messages):
        """Write all messages to the stream in a thread-safe way."""
        if not email_messages:
            return
        msg_count = 0
        with self._lock:
            try:
                stream_created = self.open()
                for message in email_messages:
                    self._process_msg(message)
                    self.write_message(message)
                    self.stream.flush()  # flush after each message
                    msg_count += 1
                if stream_created:
                    self.close()
            except Exception:
                if not self.fail_silently:
                    raise
        return msg_count



from typing import Optional

from django.utils.translation import gettext as _
from wbcore.contrib.color.enums import WBColor
from wbcore.enums import Unit
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.display import Display
from wbcore.metadata.configs.display.instance_display.layouts.inlines import Inline
from wbcore.metadata.configs.display.instance_display.layouts.layouts import Layout
from wbcore.metadata.configs.display.instance_display.operators import default
from wbcore.metadata.configs.display.instance_display.pages import Page
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    create_simple_display,
    create_simple_section,
)
from wbcore.metadata.configs.display.instance_display.styles import Style
from wbcore.metadata.configs.display.instance_display.utils import repeat_field
from wbcore.metadata.configs.display.view_config import DisplayViewConfig
from wbmailing import models


class MassMailDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="subject", label=_("Subject")),
                dp.Field(key="from_email", label=_("From")),
                dp.Field(key="template", label=_("Template")),
                dp.Field(key="mailing_lists", label=_("Mailing Lists")),
                dp.Field(key="excluded_mailing_lists", label=_("Excluded Mailing Lists")),
                dp.Field(key="created", label=_("Created")),
                dp.Field(key="creator", label=_("Creator")),
                dp.Field(key="send_at", label=_("Send At")),
            ],
            legends=[
                dp.Legend(
                    key="status",
                    items=[
                        dp.LegendItem(
                            icon=WBColor.GREEN_LIGHT.value,
                            label=models.MassMail.Status.SENT.label,
                            value=models.MassMail.Status.SENT,
                        ),
                        dp.LegendItem(
                            icon=WBColor.GREEN.value,
                            label=models.MassMail.Status.SEND_LATER.label,
                            value=models.MassMail.Status.SEND_LATER,
                        ),
                        dp.LegendItem(
                            icon=WBColor.YELLOW_LIGHT.value,
                            label=models.MassMail.Status.PENDING.label,
                            value=models.MassMail.Status.PENDING,
                        ),
                        dp.LegendItem(
                            icon=WBColor.BLUE_LIGHT.value,
                            label=models.MassMail.Status.DRAFT.label,
                            value=models.MassMail.Status.DRAFT,
                        ),
                        dp.LegendItem(
                            icon=WBColor.RED_LIGHT.value,
                            label=models.MassMail.Status.DENIED.label,
                            value=models.MassMail.Status.DENIED,
                        ),
                    ],
                )
            ],
            formatting=[
                dp.Formatting(
                    column="status",
                    formatting_rules=[
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.GREEN_LIGHT.value},
                            condition=("==", models.MassMail.Status.SENT),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.GREEN.value},
                            condition=("==", models.MassMail.Status.SEND_LATER),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.YELLOW_LIGHT.value},
                            condition=("==", models.MassMail.Status.PENDING),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.BLUE_LIGHT.value},
                            condition=("==", models.MassMail.Status.DRAFT),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.RED_LIGHT.value},
                            condition=("==", models.MassMail.Status.DENIED),
                        ),
                    ],
                )
            ],
        )

    def get_instance_display(self) -> Display:
        display = Display(
            pages=[
                Page(
                    title="Home",
                    layouts={
                        default(): Layout(
                            grid_template_areas=[
                                ["status", "status"],
                                ["subject", "subject"],
                                ["from_email", "template"],
                                ["mailing_lists", "excluded_mailing_lists"],
                                ["body", "body"],
                                ["attachment_url", "attachment_url"],
                            ]
                        )
                    },
                ),
                Page(
                    title="Analytics",
                    layouts={
                        default(): Layout(
                            grid_template_areas=[
                                ["mailstatus_barchart", "mailclick_barchart"],
                                ["region_barchart", "region_barchart"],
                                ["clients_barchart", "country_barchart"],
                            ],
                            grid_template_rows=[Style.px(600), Style.px(600), Style.px(600)],
                            inlines=[
                                Inline(key="mailstatus_barchart", endpoint="mailstatus_barchart"),
                                Inline(key="mailclick_barchart", endpoint="mailclick_barchart"),
                                Inline(key="clients_barchart", endpoint="clients_barchart"),
                                Inline(key="country_barchart", endpoint="country_barchart"),
                                Inline(key="region_barchart", endpoint="region_barchart"),
                            ],
                        )
                    },
                ),
                Page(
                    title="Events",
                    layouts={
                        default(): Layout(
                            grid_template_areas=[["events_key"]],
                            inlines=[Inline(key="events_key", endpoint="events")],
                            grid_template_columns=[
                                "minmax(min-content, 1fr)",
                            ],
                            grid_auto_rows=Style.MIN_CONTENT,
                        ),
                    },
                ),
            ]
        )

        return display


EVENTTYPE_LEGENDS = dp.Legend(
    key="status",
    items=[
        dp.LegendItem(
            icon=WBColor.GREY.value,
            label=models.MailEvent.EventType.CREATED.label,
            value=models.MailEvent.EventType.CREATED,
        ),
        dp.LegendItem(
            icon=WBColor.BLUE_LIGHT.value,
            label=models.MailEvent.EventType.QUEUED.label,
            value=models.MailEvent.EventType.QUEUED,
        ),
        dp.LegendItem(
            icon=WBColor.BLUE_DARK.value,
            label=models.MailEvent.EventType.UNKNOWN.label,
            value=models.MailEvent.EventType.UNKNOWN,
        ),
        dp.LegendItem(
            icon=WBColor.YELLOW_LIGHT.value,
            label=models.MailEvent.EventType.DELIVERED.label,
            value=models.MailEvent.EventType.DELIVERED,
        ),
        dp.LegendItem(
            icon=WBColor.GREEN_LIGHT.value,
            label=models.MailEvent.EventType.OPENED.label,
            value=models.MailEvent.EventType.OPENED,
        ),
        dp.LegendItem(
            icon=WBColor.GREEN.value,
            label=models.MailEvent.EventType.CLICKED.label,
            value=models.MailEvent.EventType.CLICKED,
        ),
        dp.LegendItem(
            icon=WBColor.RED_LIGHT.value,
            label=models.MailEvent.EventType.DEFERRED.label,
            value=models.MailEvent.EventType.DEFERRED,
        ),
        dp.LegendItem(
            icon=WBColor.RED.value,
            label=models.MailEvent.EventType.BOUNCED.label,
            value=models.MailEvent.EventType.BOUNCED,
        ),
        dp.LegendItem(
            icon=WBColor.RED_DARK.value,
            label=models.MailEvent.EventType.REJECTED.label,
            value=models.MailEvent.EventType.REJECTED,
        ),
    ],
)

EVENTTYPE_FORMATTING = dp.Formatting(
    column="status",
    formatting_rules=[
        dp.FormattingRule(
            style={"backgroundColor": WBColor.BLUE_LIGHT.value},
            condition=("==", models.MailEvent.EventType.QUEUED),
        ),
        dp.FormattingRule(
            style={"backgroundColor": WBColor.GREEN.value},
            condition=("==", models.MailEvent.EventType.CLICKED),
        ),
        dp.FormattingRule(
            style={"backgroundColor": WBColor.GREEN_LIGHT.value},
            condition=("==", models.MailEvent.EventType.OPENED),
        ),
        dp.FormattingRule(
            style={"backgroundColor": WBColor.YELLOW_LIGHT.value},
            condition=("==", models.MailEvent.EventType.DELIVERED),
        ),
        dp.FormattingRule(
            style={"backgroundColor": WBColor.RED.value},
            condition=("==", models.MailEvent.EventType.BOUNCED),
        ),
        dp.FormattingRule(
            style={"backgroundColor": WBColor.RED_LIGHT.value},
            condition=("==", models.MailEvent.EventType.DEFERRED),
        ),
        dp.FormattingRule(
            style={"backgroundColor": WBColor.GREY.value},
            condition=("==", models.MailEvent.EventType.CREATED),
        ),
        dp.FormattingRule(
            style={"backgroundColor": WBColor.RED_DARK.value},
            condition=("==", models.MailEvent.EventType.REJECTED),
        ),
    ],
)


class MailDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="mass_mail", label=_("Mass Mail")),
                dp.Field(key="subject", label=_("Subject")),
                dp.Field(key="from_email", label=_("From")),
                dp.Field(key="to_email", label=_("To")),
                dp.Field(key="last_send", label=_("Last Sent")),
                dp.Field(key="created", label=_("Created")),
                dp.Field(key="status", label=_("Latest Status")),
            ],
            legends=[EVENTTYPE_LEGENDS],
            formatting=[EVENTTYPE_FORMATTING],
        )

    def get_instance_display(self) -> Display:
        fields = [
            [repeat_field(2, "subject")],
            ["from_email", "to_email"],
            ["cc_email", "bcc_email"],
            ["created", "."],
        ]
        instance = None
        try:
            instance = self.view.get_object()
        except AssertionError:
            pass

        if instance and instance.mass_mail:
            fields.append([repeat_field(2, "mass_mail")])
        else:
            fields.append([repeat_field(2, "template")])
            fields.append([repeat_field(2, "body")])
        fields.append([repeat_field(2, "mailevent_section")])
        return create_simple_display(
            fields, [create_simple_section("mailevent_section", _("Events"), [["mailevent"]], "mailevent")]
        )


class MailStatusMassMailDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="entry", label=_("Recipient"), width=Unit.PIXEL(350)),
                dp.Field(key="address", label=_("Mail Address"), width=Unit.PIXEL(200)),
                dp.Field(key="status", label=_("Latest Status"), width=Unit.PIXEL(200)),
            ],
            legends=[EVENTTYPE_LEGENDS],
            formatting=[EVENTTYPE_FORMATTING],
        )


class MailTemplateDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(fields=[dp.Field(key="title", label=_("Title"))])

    def get_instance_display(self) -> Display:
        return create_simple_display([["title"], ["template"]])


class MailEventDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="timestamp", label=_("Date")),
                dp.Field(key="mail", label=_("Mail")) if "mail_id" not in self.view.kwargs else (),
                dp.Field(key="event_type", label=_("Type")),
                dp.Field(key="reject_reason", label=_("Rejection Reason")),
                dp.Field(key="recipient", label=_("Recipient")),
                dp.Field(key="click_url", label=_("URL")),
                dp.Field(key="ip", label=_("IP")),
                dp.Field(key="user_agent", label=_("User Agent")),
                dp.Field(key="description", label=_("Description")),
            ]
        )



from django.utils.translation import gettext as _
from wbcore.contrib.directory.models import Entry
from wbcore.metadata.configs.titles import TitleViewConfig
from wbmailing import models


class MailingListTitleConfig(TitleViewConfig):
    def get_instance_title(self):
        return _("Mailing List: {{title}}")

    def get_list_title(self):
        return _("Mailing Lists")

    def get_create_title(self):
        return _("New Mailing List")


class MailingListSubscriberChangeRequestTitleConfig(TitleViewConfig):
    def get_instance_title(self):
        return _("Mailing List Subscriber Change Request")

    def get_list_title(self):
        return _("Mailing List Subscriber Change Requests")

    def get_create_title(self):
        return _("New Mailing List Subscriber Change Request")


class EmailContactMailingListTitleConfig(TitleViewConfig):
    def get_instance_title(self):
        return _("Email Contact: {{address}}")

    def get_list_title(self):
        mailinglist = models.MailingList.objects.get(id=self.view.kwargs["mailing_list_id"])
        return _("Subscribed Emails for {}").format(mailinglist.title)


class MailMailingListChartTitleConfig(TitleViewConfig):
    def get_list_title(self):
        mailinglist = models.MailingList.objects.get(id=self.view.kwargs["mailing_list_id"])
        return _("Mail Penetration for {}").format(mailinglist.title)


class UnsubscribedEmailContactMailingListTitleConfig(EmailContactMailingListTitleConfig):
    def get_list_title(self):
        mailinglist = models.MailingList.objects.get(id=self.view.kwargs["mailing_list_id"])
        return _("Unsubscribed Emails for {}").format(mailinglist.title)


class MailingListEntryTitleConfig(MailingListTitleConfig):
    def get_list_title(self):
        entry = Entry.objects.get(id=self.view.kwargs["entry_id"])
        return _("Subscribed Mailing Lists of {}").format(entry.computed_str)


class MailingListSubscriberRequestEntryTitleConfig(MailingListSubscriberChangeRequestTitleConfig):
    def get_list_title(self):
        entry = Entry.objects.get(id=self.view.kwargs["entry_id"])
        return _("Mailing Subscription Requests for {}").format(entry.computed_str)

    def get_create_title(self):
        entry = Entry.objects.get(id=self.view.kwargs["entry_id"])
        return _("New Mailing Subscription Request for {}").format(entry.computed_str)



from unittest.mock import patch

import pytest
from django.core import mail
from rest_framework import status
from rest_framework.test import APIRequestFactory
from wbcore.contrib.directory.factories import EmailContactFactory
from wbcore.contrib.directory.viewsets import (
    CompanyModelViewSet,
    EntryModelViewSet,
    PersonModelViewSet,
)
from wbcore.signals.instance_buttons import add_instance_button
from wbcore.test.utils import get_kwargs, get_model_factory
from wbmailing.factories import (
    CustomMassMailFactory,
    MailEventFactory,
    MailFactory,
    MailingListEmailContactThroughModelFactory,
    MailingListFactory,
    MailingListSubscriberChangeRequestFactory,
    MassMailFactory,
    ToEmailMailFactory,
)
from wbmailing.models import (
    MailEvent,
    MailingListEmailContactThroughModel,
    MailingListSubscriberChangeRequest,
)
from wbmailing.viewsets import (
    EmailContactMailingListModelViewSet,
    MailingListEntryModelViewSet,
    MailingListSubscriberChangeRequestModelViewSet,
    MailModelViewSet,
    MailStatusMassMailModelViewSet,
    ManageMailingListSubscriptions,
    MassMailModelViewSet,
    UnsubscribeView,
)


@pytest.mark.django_db
class TestMailStatusMassMailViewSet:
    @pytest.fixture()
    def mass_mail(self, user):
        email = EmailContactFactory.create()
        ml = MailingListFactory.create(email_contacts=[email])
        mass_mail = MassMailFactory.create(mailing_lists=[ml], status="SENT")
        mail = MailFactory.create(mass_mail=mass_mail)
        MailEventFactory.create(mail=mail, event_type=MailEvent.EventType.BOUNCED, recipient=email.address)
        return mass_mail

    @patch("wbmailing.models.mails.send_mail_as_task.delay")
    @pytest.mark.parametrize("mvs", [MailStatusMassMailModelViewSet])
    def test_resendbouncedmails(self, send_mail_as_task, mvs, mass_mail, user):
        request = APIRequestFactory().get("")
        request.user = user
        kwargs = {"mass_mail_id": mass_mail.id}

        response = mvs(kwargs=kwargs).resendbouncedmails(request=request, mass_mail_id=mass_mail.id)
        assert response.status_code == status.HTTP_200_OK
        assert response.data
        send_mail_as_task.assert_called()
        assert send_mail_as_task.call_count == 1

    @pytest.mark.parametrize("mvs", [MailStatusMassMailModelViewSet])
    def test_get_custom_buttons(self, mvs, mass_mail, user):
        request = APIRequestFactory().get("")
        request.user = user
        # obj = mail_factory()
        kwargs = {"mass_mail_id": mass_mail.id}
        mvs.kwargs = kwargs
        response = (
            mvs(kwargs=kwargs).button_config_class(view=mvs, request=request, instance=mass_mail).get_custom_buttons()
        )
        assert response


@pytest.mark.django_db
class TestSpecificViewsets:
    @pytest.mark.parametrize(
        "mvs, factory",
        [
            (EmailContactMailingListModelViewSet, MailingListEmailContactThroughModelFactory),
        ],
    )
    def test_delete(self, mvs, factory, user):
        request = APIRequestFactory().get("")
        request.user = user
        obj = factory()
        kwargs = get_kwargs(obj, mvs, request)
        response = mvs(kwargs=kwargs).delete(request=request, mailing_list_id=kwargs["mailing_list_id"], pk=obj.pk)
        assert response.status_code == status.HTTP_200_OK
        assert response.data

    @pytest.mark.parametrize(
        "mvs, factory", [(MailingListEntryModelViewSet, MailingListEmailContactThroughModelFactory)]
    )
    def test_delete_unsubscriber_email_contacts(self, mvs, factory, mailing_list_factory, user):
        request = APIRequestFactory().get("")
        request.user = user
        mlscr = MailingListSubscriberChangeRequestFactory()
        obj = factory()
        kwargs = get_kwargs(obj, mvs, request)
        response = mvs(kwargs=kwargs).unsubscribe(
            request=request, entry_id=mlscr.email_contact.entry.id, pk=mlscr.mailing_list.pk
        )
        assert response.status_code == status.HTTP_200_OK
        assert response.data

    @pytest.mark.parametrize("myview", [PersonModelViewSet, CompanyModelViewSet, EntryModelViewSet])
    def test_entry_adding_instance_buttons(self, myview, user):
        remote_buttons = add_instance_button.send(myview, many=True)
        custom_instance_buttons = set([button for _, button in remote_buttons])
        assert custom_instance_buttons

    # @pytest.mark.parametrize("mvs", [(MassMailModelViewSet)])
    # def test_get_messages(self, mvs, mass_mail_factory, mail_factory, mail_event_factory, user):
    #     request = APIRequestFactory().get("")
    #     request.user = user
    #     obj = mass_mail_factory(status = MassMail.Status.SENT)
    #     qs = mvs().get_queryset()
    #     mail = mail_factory(mass_mail=qs[0])
    #     me = mail_event_factory(mail = mail, event_type=MailEvent.EventType.CLICKED)

    #     msg = mvs().get_messages(request, instance=qs[0])
    #     assert msg

    @pytest.mark.parametrize("mvs, factory", [(MassMailModelViewSet, CustomMassMailFactory)])
    def test_sendtestmail(self, mvs, factory, user):
        request = APIRequestFactory().get("")
        request.user = user
        obj = factory()
        EmailContactFactory(address=request.user.email)
        # request.POST = request.POST.copy()
        # request.POST['to_test_email'] = "lemissan@atonra.ch"
        nb_mail_send = len(mail.outbox)
        request.query_params = {}
        mvs.request = request
        mvs.kwargs = get_kwargs(obj, mvs, request)
        mvs.kwargs["pk"] = obj.pk
        response = mvs().sendtestmail(request)
        assert response.status_code == status.HTTP_200_OK
        assert response.data
        assert len(mail.outbox) == nb_mail_send + 1

    @pytest.mark.parametrize("mvs, factory", [(MailModelViewSet, ToEmailMailFactory)])
    def test_get_queryset_mail(self, mvs, factory, user):
        request = APIRequestFactory().get("")
        request.user = user
        request.user.is_superuser = False
        request.user.save()
        obj = factory()
        kwargs = get_kwargs(obj, mvs, request)
        vs = mvs.as_view({"get": "list"})
        response = vs(request, **kwargs)
        assert response.status_code == status.HTTP_403_FORBIDDEN

    @patch("wbmailing.models.mails.send_mail_as_task.delay")
    def test_resend(self, send_mail_as_task, user):
        request = APIRequestFactory().get("")
        request.user = user
        obj = ToEmailMailFactory()
        request.query_params = {}
        MailModelViewSet.request = request
        MailModelViewSet.kwargs = get_kwargs(obj, MailModelViewSet, request)
        MailModelViewSet.kwargs["pk"] = obj.pk
        response = MailModelViewSet().resend(request)
        assert response.status_code == status.HTTP_200_OK
        assert response.data
        send_mail_as_task.assert_called()
        assert send_mail_as_task.call_count == 1

    @pytest.mark.parametrize("user__is_superuser", [True])
    def test_get_custom_buttons_approveall(self, user):
        request = APIRequestFactory().get("")
        request.user = user
        factory = get_model_factory(MailingListSubscriberChangeRequestModelViewSet().get_serializer_class().Meta.model)
        for i in range(2):
            obj = factory()
        kwargs = get_kwargs(obj, MailingListSubscriberChangeRequestModelViewSet, request)
        assert (
            MailingListSubscriberChangeRequest.objects.filter(
                status=MailingListSubscriberChangeRequest.Status.PENDING
            ).count()
            == 2
        )
        response = (
            MailingListSubscriberChangeRequestModelViewSet(kwargs=kwargs)
            .button_config_class(view=MailingListSubscriberChangeRequestModelViewSet, request=request, instance=obj)
            .get_custom_buttons()
        )
        assert response

    @pytest.mark.parametrize(
        "mvs, factory", [(ManageMailingListSubscriptions, MailingListEmailContactThroughModelFactory)]
    )
    def test_ManageMailingListSubscriptions(self, mvs, factory, user):
        request = APIRequestFactory().get("")
        request.user = user
        through_model = factory()
        response = mvs().get(request, through_model.email_contact.id)
        assert response.status_code == status.HTTP_200_OK
        assert response.content

    @pytest.mark.parametrize("mvs, factory", [(UnsubscribeView, MailingListEmailContactThroughModelFactory)])
    def test_UnsubscribeView(self, mvs, factory, user):
        request = APIRequestFactory().get("")
        request.user = user
        through_model = factory()
        mailing_list = through_model.mailing_list
        response = mvs().get(request, through_model.email_contact.id, mailing_list.id)
        assert response.status_code == status.HTTP_302_FOUND
        assert response.url
        through_model.refresh_from_db()
        assert through_model.status == MailingListEmailContactThroughModel.Status.UNSUBSCRIBED
